from .api import FlickrApi
from .downloader import download_file
from .exceptions import (
    FlickrApiException,
    InsufficientPermissionsToComment,
    InvalidApiKey,
    InvalidXmlException,
    PhotoIsPrivate,
    ResourceNotFound,
    LicenseNotFound,
    UnrecognisedFlickrApiException,
    UserDeleted,
)
from .types import (
    Comment,
    CommonsInstitution,
    MachineTags,
    License,
    User,
    UserInfo,
    TakenGranularity,
    DateTaken,
    LocationInfo,
    Size,
    SafetyLevel,
    SinglePhoto,
    CollectionOfPhotos,
    PhotosFromUrl,
    PhotosInAlbum,
    PhotosInGallery,
    PhotosInGroup,
    AlbumContext,
    AlbumInfo,
    GalleryContext,
    GalleryInfo,
    GroupContext,
    GroupInfo,
    MediaType,
    PhotoContext,
    SinglePhotoInfo,
    Tag,
    Visibility,
)


__version__ = "2.17.2"


__all__ = [
    "download_file",
    "FlickrApi",
    "FlickrApiException",
    "ResourceNotFound",
    "InvalidApiKey",
    "InvalidXmlException",
    "LicenseNotFound",
    "License",
    "LocationInfo",
    "MachineTags",
    "User",
    "UserInfo",
    "TakenGranularity",
    "Comment",
    "CommonsInstitution",
    "DateTaken",
    "Size",
    "SafetyLevel",
    "SinglePhoto",
    "SinglePhotoInfo",
    "CollectionOfPhotos",
    "PhotosFromUrl",
    "PhotosInAlbum",
    "PhotosInGallery",
    "PhotosInGroup",
    "AlbumContext",
    "AlbumInfo",
    "GalleryContext",
    "GalleryInfo",
    "GroupContext",
    "GroupInfo",
    "InsufficientPermissionsToComment",
    "MediaType",
    "PhotoContext",
    "PhotoIsPrivate",
    "Tag",
    "UnrecognisedFlickrApiException",
    "UserDeleted",
    "Visibility",
    "__version__",
]



"""
Convert values retrieved from the Flickr API into nicely-typed values.
"""

from xml.etree import ElementTree as ET

from .types import LocationInfo, SafetyLevel, Size


def parse_safety_level(s: str) -> SafetyLevel:
    """
    Converts a numeric safety level ID in the Flickr API into
    a human-readable value.

    See https://www.flickrhelp.com/hc/en-us/articles/4404064206996-Content-filters
    """
    lookup_table: dict[str, SafetyLevel] = {
        "0": "safe",
        "1": "moderate",
        "2": "restricted",
    }

    try:
        return lookup_table[s]
    except KeyError:
        raise ValueError(f"Unrecognised safety level: {s}")


def parse_sizes(photo_elem: ET.Element) -> list[Size]:
    """
    Get a list of sizes from a photo in a collection response.
    """
    # When you get a collection of photos (e.g. in an album)
    # you can get some of the sizes on the <photo> element, e.g.
    #
    #     <
    #       photo
    #       url_t="https://live.staticflickr.com/2893/1234567890_t.jpg"
    #       height_t="78"
    #       width_t="100"
    #       …
    #     />
    #
    sizes: list[Size] = []

    for suffix, label in [
        ("sq", "Square"),
        ("q", "Large Square"),
        ("t", "Thumbnail"),
        ("s", "Small"),
        ("m", "Medium"),
        ("l", "Large"),
        ("o", "Original"),
    ]:
        try:
            media = photo_elem.attrib["media"]

            if media not in ("video", "photo"):  # pragma: no cover
                raise ValueError(f"Unrecognised media: {media!r}")

            sizes.append(
                {
                    "height": int(photo_elem.attrib[f"height_{suffix}"]),
                    "width": int(photo_elem.attrib[f"width_{suffix}"]),
                    "label": label,
                    "media": media,  # type: ignore
                    "source": photo_elem.attrib[f"url_{suffix}"],
                }
            )
        except KeyError:
            pass

    return sizes


def parse_location(elem_with_location: ET.Element) -> LocationInfo | None:
    """
    Get location information about a photo.

    This takes an XML element with latitude/longitude/accuracy attributes;
    this can be a <location> element (on a single photo) or a <photo> element
    (on collection responses).
    """
    # The accuracy parameter in the Flickr API response tells us
    # the precision of the location information (15 November 2023):
    #
    #     Recorded accuracy level of the location information.
    #     World level is 1, Country is ~3, Region ~6, City ~11, Street ~16.
    #     Current range is 1-16.
    #
    # But some photos have an accuracy of 0!  It's unclear what this
    # means or how we should map this -- lower numbers mean less accurate,
    # so this location information might be completely useless.
    #
    # Discard it rather than risk propagating bad data.
    if elem_with_location.attrib["accuracy"] == "0":
        return None

    return {
        "latitude": float(elem_with_location.attrib["latitude"]),
        "longitude": float(elem_with_location.attrib["longitude"]),
        "accuracy": int(elem_with_location.attrib["accuracy"]),
    }



"""
Tests for ``flickr_photos_api.utils``.
"""

import pytest

from flickr_photos_api.utils import parse_safety_level


def test_unrecognised_safety_level_is_error() -> None:
    """
    Parsing an unrecognised value as a safety level throws ``ValueError``.
    """
    with pytest.raises(ValueError, match="Unrecognised safety level"):
        parse_safety_level("-1")



"""
Tests for ``flickr_photos_api.date_parsers``.
"""

import pytest

from flickr_photos_api.date_parsers import parse_date_taken


def test_unrecognised_date_granularity_is_error() -> None:
    """
    If you pass an unrecognised date taken granularity,
    it throws ``ValueError``.
    """
    with pytest.raises(ValueError, match="Unrecognised date granularity"):
        parse_date_taken(value="2017-02-17 00:00:00", granularity="-1", unknown=False)



"""
The Flickr API returns dates in several formats; the helpers in this file
parse these dates and turn them into native Python types.
"""

import datetime

from flickr_photos_api.types import DateTaken, TakenGranularity


def parse_timestamp(ts: str, /) -> datetime.datetime:
    """
    Convert a Unix timestamp into a Python-native ``datetime.datetime``.

    Example:

        >>> parse_timestamp('1490376472')
        datetime.datetime(2017, 3, 24, 17, 27, 52, tzinfo=datetime.timezone.utc)

    The Flickr API frequently returns dates as Unix timestamps, for example:

    *   When you call ``flickr.photos.getInfo``, the ``<dates>`` element
        includes the upload and last update dates as a timestamp
    *   When you call ``flickr.people.getInfo`` for a user with Flickr Pro,
        the ``expires`` attribute is a numeric timestamp.

    In this case a Unix timestamp is "an unsigned integer specifying
    the number of seconds since Jan 1st 1970 GMT" [1].

    [1] https://www.flickr.com/services/api/misc.dates.html
    """
    return datetime.datetime.fromtimestamp(int(ts), tz=datetime.timezone.utc)


def _parse_date_taken_value(dt: str) -> datetime.datetime:
    """
    Convert a "date taken" string to a Python-native ``datetime.datetime``.

    Example:

        >>> _parse_date_taken_value('2017-02-17 00:00:00')
        datetime.datetime(2017, 2, 17, 0, 0)

    """
    # See https://www.flickr.com/services/api/misc.dates.html
    #
    #     The date taken should always be displayed in the timezone
    #     of the photo owner, which is to say, don't perform
    #     any conversion on it.
    #
    return datetime.datetime.strptime(dt, "%Y-%m-%d %H:%M:%S")


def _parse_date_taken_granularity(g: str) -> TakenGranularity:
    """
    Converts a numeric granularity level in the Flickr API into a
    human-readable value.

    See https://www.flickr.com/services/api/misc.dates.html
    """
    lookup_table: dict[str, TakenGranularity] = {
        "0": "second",
        "4": "month",
        "6": "year",
        "8": "circa",
    }

    try:
        return lookup_table[g]
    except KeyError:
        raise ValueError(f"Unrecognised date granularity: {g}")


def parse_date_taken(
    *, value: str, granularity: str, unknown: bool
) -> DateTaken | None:
    """
    Parse a "date taken" value from the Flickr API.

    When you retrieve a photo with ``flickr.photos.getInfo``, the taken
    value is made of several parts:

        <dates … takengranularity="0" takenunknown="1" lastupdate="1705439827"/>

    This function converts these parts into nicely-typed objects.  This
    may be ``None`` if the date taken is unknown.
    """
    # We intentionally omit sending any 'date taken' information
    # to callers if it's unknown.
    #
    # There will be a value in the API response, but if the taken date
    # is unknown, it's defaulted to the date the photo was posted.
    # See https://www.flickr.com/services/api/misc.dates.html
    #
    # This value isn't helpful to callers, so we omit it.  This reduces
    # the risk of somebody skipping the ``unknown`` parameter and using
    # the value in the wrong place.
    if unknown:
        return None

    # This is a weird value I've seen returned on some videos; I'm
    # not sure what it means, but it's not something we can interpret
    # as a valid date, so we treat "date taken" as unknown even if
    # the API thinks it knows it.
    elif value.startswith("0000-"):
        return None

    else:
        return {
            "value": _parse_date_taken_value(value),
            "granularity": _parse_date_taken_granularity(granularity),
        }



"""
For Flickr, a "context" is anywhere a photo appears as part of
a larger collection -- not just on its own.

It could be part of an album, a gallery, or a group.
"""

import datetime
import typing

from .users import User


class AlbumContext(typing.TypedDict):
    """
    An album is a collection of photos that are owned by the
    album's creator.

    If Alice creates an album, she can put her photos in it, but
    she can't put Bob's photos in her album, and vice versa.
    """

    id: str
    title: str

    count_photos: int
    count_videos: int

    count_views: int
    count_comments: int


class GalleryContext(typing.TypedDict):
    """
    An gallery is a collection of photos that are not owned by
    the album's gallery's photos.

    If Alice creates an album, she can't put her own photos in it,
    but she could put Bob's photos in it.
    """

    id: str
    url: str
    owner: User

    title: str
    description: str | None

    date_created: datetime.datetime
    date_updated: datetime.datetime

    count_photos: int
    count_videos: int

    count_views: int
    count_comments: int


class GroupContext(typing.TypedDict):
    """
    An group is a collection of Flickr users with a common interest,
    who create a shared pool of relevant photos.

    Alice could create a group and add her photos to the group pool.
    Bob could then join as a member and add some of his photos to
    the pool.
    """

    id: str
    title: str
    url: str

    count_items: int
    count_members: int


class PhotoContext(typing.TypedDict):
    """
    Places where a photo might appear on Flickr.com.
    """

    albums: list[AlbumContext]
    galleries: list[GalleryContext]
    groups: list[GroupContext]



"""
Methods for getting information about a single photo in the Flickr API.
"""

import typing
from xml.etree import ElementTree as ET

from flickr_url_parser import looks_like_flickr_photo_id
from nitrate.xml import find_optional_text, find_required_elem, find_required_text

from flickr_photos_api.date_parsers import parse_date_taken, parse_timestamp
from .license_methods import LicenseMethods
from ..exceptions import PhotoIsPrivate, ResourceNotFound
from ..types import (
    AlbumContext,
    GalleryContext,
    GroupContext,
    MediaType,
    PhotoContext,
    SinglePhotoInfo,
    SinglePhoto,
    Size,
    Tag,
    Visibility,
    create_user,
    get_machine_tags,
)
from ..utils import parse_location, parse_safety_level


class SinglePhotoMethods(LicenseMethods):
    """
    Methods for getting information about a single photo.
    """

    def parse_single_photo_info(
        self, info_resp: ET.Element, *, photo_id: str
    ) -> SinglePhotoInfo:
        """
        Parse the XML response from the ``flickr.photos.getInfo`` API.
        """
        # The getInfo response is a blob of XML of the form:
        #
        #       <?xml version="1.0" encoding="utf-8" ?>
        #       <rsp stat="ok">
        #       <photo license="8" …>
        #           <owner
        #               nsid="30884892@N08
        #               username="U.S. Coast Guard"
        #               realname="Coast Guard" …
        #           >
        #               …
        #           </owner>
        #           <title>Puppy Kisses</title>
        #           <description>Seaman Nina Bowen shows …</description>
        #           <dates
        #               posted="1490376472"
        #               taken="2017-02-17 00:00:00"
        #               …
        #           />
        #           <urls>
        #               <url type="photopage">https://www.flickr.com/photos/coast_guard/32812033543/</url>
        #           </urls>
        #           <tags>
        #           <tag raw="indian ocean" …>indianocean</tag>
        #           …
        #       </photo>
        #       </rsp>
        #
        photo_elem = find_required_elem(info_resp, path=".//photo")

        safety_level = parse_safety_level(photo_elem.attrib["safety_level"])

        license = self.lookup_license_by_id(id=photo_elem.attrib["license"])

        title = find_optional_text(photo_elem, path="title")
        description = find_optional_text(photo_elem, path="description")

        owner_elem = find_required_elem(photo_elem, path="owner")
        owner = create_user(
            user_id=owner_elem.attrib["nsid"],
            username=owner_elem.attrib["username"],
            realname=owner_elem.attrib["realname"],
            path_alias=owner_elem.attrib["path_alias"],
        )

        dates = find_required_elem(photo_elem, path="dates").attrib

        date_posted = parse_timestamp(dates["posted"])

        date_taken = parse_date_taken(
            value=dates["taken"],
            granularity=dates["takengranularity"],
            unknown=dates["takenunknown"] == "1",
        )

        url = find_required_text(photo_elem, path='.//urls/url[@type="photopage"]')

        count_comments = int(find_required_text(photo_elem, path="comments"))
        count_views = int(photo_elem.attrib["views"])

        # The originalformat parameter will only be returned if the user
        # allows downloads of the photo.
        #
        # We only need this parameter for photos that can be uploaded to
        # Wikimedia Commons.  All CC-licensed photos allow downloads, so
        # we'll always get this parameter for those photos.
        #
        # See https://www.flickr.com/help/forum/32218/
        # See https://www.flickrhelp.com/hc/en-us/articles/4404079715220-Download-permissions
        original_format = photo_elem.get("originalformat")

        # When you look up tags on a single photo, you get a structured
        # block of data, e.g.:
        #
        #     <tags>
        #       <tag
        #         id="32695993-21609597615-765"
        #         author="126912357@N06"
        #         authorname="valuable basket"
        #         raw="Church"
        #         machine_tag="0">church</tag>
        #       ...
        #     </tags>
        #
        # We write the normalized value into the "tags" field, because
        # this matches the values we get from the collection endpoints.
        #
        # However, in some context it's useful to have the raw value as
        # entered by the user, so we store this in the ``raw_tags`` list.
        #
        # Note: it's rare, but some Flickr photos do have empty text for
        # the tag value.  See the tests for an example.
        tags_elem = find_required_elem(photo_elem, path="tags")

        tags = []
        raw_tags: list[Tag] = []

        for t in tags_elem.findall("tag"):
            tags.append(t.text or "")
            raw_tags.append(
                {
                    "author_id": t.attrib["author"],
                    "author_name": t.attrib["authorname"],
                    "raw_value": t.attrib["raw"],
                    "normalized_value": t.text or "",
                    "is_machine_tag": t.attrib["machine_tag"] == "1",
                }
            )

        # Get location information about the photo.
        #
        # The <location> tag is only present in photos which have
        # location data; if the user hasn't made location available to
        # public users, it'll be missing.
        location_elem = photo_elem.find(path="location")

        if location_elem is not None:
            location = parse_location(location_elem)
        else:
            location = None

        # Get visibility information about the photo.
        #
        # This is returned in the form:
        #
        #     <visibility ispublic="1" isfriend="0" isfamily="0"/>
        #
        visibility_elem = find_required_elem(photo_elem, path="visibility")
        visibility: Visibility = {
            "is_public": visibility_elem.attrib["ispublic"] == "1",
            "is_friend": visibility_elem.attrib["isfriend"] == "1",
            "is_family": visibility_elem.attrib["isfamily"] == "1",
        }

        assert photo_elem.attrib["media"] in {"photo", "video"}
        media_type = typing.cast(MediaType, photo_elem.attrib["media"])

        return {
            "id": photo_id,
            "media": media_type,
            "secret": photo_elem.attrib["secret"],
            "server": photo_elem.attrib["server"],
            "farm": photo_elem.attrib["farm"],
            "original_format": original_format,
            "owner": owner,
            "safety_level": safety_level,
            "license": license,
            "title": title,
            "description": description,
            "tags": tags,
            "raw_tags": raw_tags,
            "machine_tags": get_machine_tags(tags),
            "date_posted": date_posted,
            "date_taken": date_taken,
            "location": location,
            "count_comments": count_comments,
            "count_views": count_views,
            "url": url,
            "visibility": visibility,
        }

    def get_single_photo_info(self, *, photo_id: str) -> SinglePhotoInfo:
        """
        Look up the information for a single photo.

        This uses the flickr.photos.getInfo API.
        """
        if not looks_like_flickr_photo_id(photo_id):
            raise ValueError(f"Not a Flickr photo ID: {photo_id!r}")

        info_resp = self.call(
            method="flickr.photos.getInfo",
            params={"photo_id": photo_id},
            exceptions={
                "1": ResourceNotFound(f"Could not find photo with ID: {photo_id!r}"),
                "2": PhotoIsPrivate(photo_id),
            },
        )

        return self.parse_single_photo_info(info_resp, photo_id=photo_id)

    def get_single_photo_sizes(self, *, photo_id: str) -> list[Size]:
        """
        Look up the sizes for a single photo.

        This uses the flickr.photos.getSizes API.
        """
        sizes_resp = self.call(
            method="flickr.photos.getSizes", params={"photo_id": photo_id}
        )

        # The getSizes response is a blob of XML of the form:
        #
        #       <?xml version="1.0" encoding="utf-8" ?>
        #       <rsp stat="ok">
        #       <sizes canblog="0" canprint="0" candownload="1">
        #           <size
        #               label="Square"
        #               width="75"
        #               height="75"
        #               source="https://live.staticflickr.com/2903/32812033543_c1b3784192_s.jpg"
        #               url="https://www.flickr.com/photos/coast_guard/32812033543/sizes/sq/"
        #               media="photo"
        #           />
        #           <size
        #               label="Large Square"
        #               width="150"
        #               height="150"
        #               source="https://live.staticflickr.com/2903/32812033543_c1b3784192_q.jpg"
        #               url="https://www.flickr.com/photos/coast_guard/32812033543/sizes/q/"
        #               media="photo"
        #           />
        #           …
        #       </sizes>
        #       </rsp>
        #
        # Within this function, we just return all the sizes -- we leave it up to the
        # caller to decide which size is most appropriate for their purposes.
        sizes: list[Size] = []

        for s in sizes_resp.findall(".//size"):
            if s.attrib["media"] == "photo":
                sizes.append(
                    {
                        "label": s.attrib["label"],
                        "width": int(s.attrib["width"]),
                        "height": int(s.attrib["height"]),
                        "media": "photo",
                        "source": s.attrib["source"],
                    }
                )

            elif s.attrib["media"] == "video":
                try:
                    width = int(s.attrib["width"])
                except ValueError:
                    width = None

                try:
                    height = int(s.attrib["height"])
                except ValueError:
                    height = None

                sizes.append(
                    {
                        "label": s.attrib["label"],
                        "width": width,
                        "height": height,
                        "media": "video",
                        "source": s.attrib["source"],
                    }
                )
            else:  # pragma: no cover
                raise ValueError(f"Unrecognised media type: {s.attrib['media']}")

        return sizes

    def get_single_photo(self, *, photo_id: str) -> SinglePhoto:
        """
        Look up the information for a single photo.
        """
        info = self.get_single_photo_info(photo_id=photo_id)
        sizes = self.get_single_photo_sizes(photo_id=photo_id)

        return {**info, "sizes": sizes}

    def is_photo_deleted(self, *, photo_id: str) -> bool:
        """
        Check if a photo has been deleted from Flickr.

        TODO: This can tell us whether a photo exists, but not whether
        it's been deleted -- we can't know whether a given photo ID
        never existed, or if it used to exist but doesn't.

        We should change this method to ``get_photo_state()`` returning
        an enum like exists/does_not_exist/private.
        """
        try:
            self.call(
                method="flickr.photos.getInfo",
                params={"photo_id": photo_id},
                exceptions={"1": ResourceNotFound(), "2": PhotoIsPrivate(photo_id)},
            )
        except ResourceNotFound:
            return True
        except PhotoIsPrivate:
            return False
        else:
            return False

    def get_photo_contexts(self, *, photo_id: str) -> PhotoContext:
        """
        Find the contexts where this photo appears on Flickr.

        This includes albums, galleries, and groups.
        """
        if not looks_like_flickr_photo_id(photo_id):
            raise ValueError(f"Not a Flickr photo ID: {photo_id!r}")

        # See https://www.flickr.com/services/api/flickr.photos.getAllContexts.html
        contexts_resp = self.call(
            method="flickr.photos.getAllContexts",
            params={"photo_id": photo_id},
            exceptions={
                "1": ResourceNotFound(f"Could not find photo with ID: {photo_id!r}")
            },
        )

        # Within the response, the albums are in XML with the following structure:
        #
        #     <
        #       set
        #       title="Landscape / Nature"
        #       id="72157650910758151"
        #       view_count="2312"
        #       comment_count="0"
        #       count_photo="269"
        #       count_video="0" […] />
        #
        albums: list[AlbumContext] = [
            {
                "id": set_elem.attrib["id"],
                "title": set_elem.attrib["title"],
                "count_photos": int(set_elem.attrib["count_photo"]),
                "count_videos": int(set_elem.attrib["count_video"]),
                "count_views": int(set_elem.attrib["view_count"]),
                "count_comments": int(set_elem.attrib["comment_count"]),
            }
            for set_elem in contexts_resp.findall(".//set")
        ]

        # Within the response, the groups are in XML with the following structure:
        #
        #     <
        #       pool
        #       title="A Picture, A Story, A Pearl"
        #       url="/groups/14776652@N22/pool/"
        #       id="14776652@N22"
        #       members="3444"
        #       pool_count="59875" […] />
        #
        groups: list[GroupContext] = [
            {
                "id": pool_elem.attrib["id"],
                "title": pool_elem.attrib["title"],
                "url": "https://www.flickr.com" + pool_elem.attrib["url"],
                "count_items": int(pool_elem.attrib["pool_count"]),
                "count_members": int(pool_elem.attrib["members"]),
            }
            for pool_elem in contexts_resp.findall(".//pool")
        ]

        # See https://www.flickr.com/services/api/flickr.galleries.getListForPhoto.html
        galleries_resp = self.call(
            method="flickr.galleries.getListForPhoto",
            params={"photo_id": photo_id, "per_page": "500"},
        )

        # Within the response, the galleries are in XML with the following structure:
        #
        #
        #     <galleries page="1" pages="1" […]>
        #       <gallery
        #         gallery_id="72157721626742458"
        #         url="https://www.flickr.com/photos/72804335@N03/galleries/72157721626742458"
        #         owner="72804335@N03"
        #         username="Josep M.Toset"
        #         date_create="1680980061"
        #         date_update="1714564015"
        #         count_photos="166"
        #         count_videos="0"
        #         count_views="152"
        #         count_comments="4" […]
        #       >
        #         <title>paisatges</title>
        #         <description/>
        #       </gallery>
        #
        galleries_elem = find_required_elem(galleries_resp, path="galleries")

        # TODO: Add supporting for paginating through the list of galleries.
        #
        # I'm not sure there are any photos where this is actually necessary,
        # so for now just throw and we can return to this if we actually see
        # a photo where this is necessary.
        if int(galleries_elem.attrib["pages"]) > 1:  # pragma: no cover
            raise ValueError("Fetching more than one page of galleries is unsupported!")

        galleries: list[GalleryContext] = []

        for gallery_elem in galleries_elem.findall("gallery"):
            gallery_url = gallery_elem.attrib["url"]

            owner_id = gallery_elem.attrib["owner"]

            path_alias = (
                gallery_url.split("/")[4]
                if gallery_url.split("/")[4] != owner_id
                else None
            )

            owner = create_user(
                user_id=owner_id,
                username=gallery_elem.attrib["username"],
                path_alias=path_alias,
                # This doesn't seem to be returned in the <gallery>
                # element even when the user has one set, so we just
                # have to accept we can't set one here.
                realname=None,
            )

            description_elem = find_required_elem(gallery_elem, path="description")
            description = description_elem.text

            galleries.append(
                {
                    "id": gallery_elem.attrib["gallery_id"],
                    "url": gallery_elem.attrib["url"],
                    "owner": owner,
                    "title": find_required_text(gallery_elem, path="title"),
                    "description": description,
                    "date_created": parse_timestamp(gallery_elem.attrib["date_create"]),
                    "date_updated": parse_timestamp(gallery_elem.attrib["date_update"]),
                    "count_photos": int(gallery_elem.attrib["count_photos"]),
                    "count_videos": int(gallery_elem.attrib["count_videos"]),
                    "count_views": int(gallery_elem.attrib["count_views"]),
                    "count_comments": int(gallery_elem.attrib["count_comments"]),
                }
            )

        return {
            "albums": albums,
            "galleries": galleries,
            "groups": groups,
        }



"""
Methods for getting information about licenses from the Flickr API.
"""

import functools

from .base import FlickrApi
from ..exceptions import LicenseNotFound
from ..types import License


class LicenseMethods(FlickrApi):
    """
    License-related methods for the Flickr API.
    """

    # Note: this list of licenses almost never changes, so we call this once
    # and cache the result for efficiency.
    @functools.cache
    def get_licenses(self) -> dict[str, License]:
        """
        Returns a list of licenses, arranged by code.

        See https://www.flickr.com/services/api/flickr.photos.licenses.getInfo.htm
        """
        license_resp = self.call(method="flickr.photos.licenses.getInfo")

        result: dict[str, License] = {}

        # Add a short ID which can be used to more easily refer to this
        # license throughout the codebase.
        license_ids = {
            "All Rights Reserved": "all-rights-reserved",
            "Attribution-NonCommercial-ShareAlike License": "cc-by-nc-sa-2.0",
            "Attribution-NonCommercial License": "cc-by-nc-2.0",
            "Attribution-NonCommercial-NoDerivs License": "cc-by-nc-nd-2.0",
            "Attribution License": "cc-by-2.0",
            "Attribution-ShareAlike License": "cc-by-sa-2.0",
            "Attribution-NoDerivs License": "cc-by-nd-2.0",
            "No known copyright restrictions": "nkcr",
            "United States Government Work": "usgov",
            "Public Domain Dedication (CC0)": "cc0-1.0",
            "Public Domain Mark": "pdm",
        }

        license_labels = {
            "Attribution-NonCommercial-ShareAlike License": "CC BY-NC-SA 2.0",
            "Attribution-NonCommercial License": "CC BY-NC 2.0",
            "Attribution-NonCommercial-NoDerivs License": "CC BY-NC-ND 2.0",
            "Attribution License": "CC BY 2.0",
            "Attribution-ShareAlike License": "CC BY-SA 2.0",
            "Attribution-NoDerivs License": "CC BY-ND 2.0",
            "Public Domain Dedication (CC0)": "CC0 1.0",
        }

        for lic in license_resp.findall(".//license"):
            result[lic.attrib["id"]] = {
                "id": license_ids[lic.attrib["name"]],
                "label": license_labels.get(lic.attrib["name"], lic.attrib["name"]),
                "url": lic.attrib["url"] or None,
            }

        return result

    @functools.cache
    def lookup_license_by_id(self, *, id: str) -> License:
        """
        Given a license ID from the Flickr API, return the license data.

        e.g. a Flickr API response might include a photo in the following form:

                <photo license="0" …>

        Then you'd call this function to find out what that means:

                >>> api.lookup_license_by_id(id="0")
                {"id": "all-rights-reserved", "name": "All Rights Reserved", "url": None}

        See https://www.flickr.com/services/api/flickr.photos.licenses.getInfo.htm
        """
        licenses = self.get_licenses()

        try:
            return licenses[id]
        except KeyError:
            raise LicenseNotFound(license_id=id)



"""
Types related to Flickr user accounts.
"""

import datetime
import typing


class User(typing.TypedDict):
    """
    Basic information about a user.  This is enough to identify the
    owner of a photo and display an attribution link.
    """

    id: str
    username: str
    realname: str | None
    path_alias: str | None
    photos_url: str
    profile_url: str


class UserInfo(User):
    """
    A user with extra information, as returned from ``user_info``.

    This includes:

    * their profile description (if any)
    * their location (if any)
    * the number of photos they've uploaded
    * their buddy icon URL
    * whether they have Flickr Pro, and if so, when it expires

    The ``pro_account_expires`` field will be present if and only if
    the user has a Pro account.
    """

    description: str | None
    location: str | None
    count_photos: int
    buddy_icon_url: str

    # Note: I tried expressing the relationship ``pro_account_expires``
    # requires ``has_pro_account = True`` in the type system, but
    # it required a Union type that broke subclassing, i.e.
    #
    #     class UserInfoWithPro(…)
    #         has_pro_account: typing.Literal[True]
    #         pro_account_expires: datetime.datetime
    #
    #     class UserInfoWithoutPro(…)
    #         has_pro_account: typing.Literal[False]
    #
    #     UserInfo = UserInfoWithPro | UserInfoWithoutPro
    #
    # but this prevents subclassing this type, which we do in the
    # Commons Explorer -- so it's a convention rather than enforced
    # by the type system.
    has_pro_account: bool
    pro_account_expires: typing.NotRequired[datetime.datetime]


def create_user(
    user_id: str, username: str, realname: str | None, path_alias: str | None
) -> User:
    """
    Given some core attributes, construct a ``User`` object.

    This function is only intended for internal user.
    """
    realname = fix_realname(user_id, username=username, realname=realname)

    # The Flickr API is a bit inconsistent about how some undefined attributes
    # are returned, e.g. ``realname`` can sometimes be null, sometimes an
    # empty string.
    #
    # In our type system, we want all of these empty values to map to ``None``.
    return {
        "id": user_id,
        "username": username,
        "realname": realname,
        "path_alias": path_alias or None,
        "photos_url": f"https://www.flickr.com/photos/{path_alias or user_id}/",
        "profile_url": f"https://www.flickr.com/people/{path_alias or user_id}/",
    }


def fix_realname(user_id: str, username: str, realname: str | None) -> str | None:
    """
    Override the ``realname`` returned by the Flickr API.

    In general we should avoid adding too many fixes here because it would
    quickly get unwieldy, but it's a useful place to consolidate these
    fixes for members we work with a lot.
    """
    realname = realname or None

    # The museum removed the 'S' so it would look good with the big 'S'
    # in their buddy icon, but that doesn't work outside Flickr.com.
    #
    # This name needed fixing on 23 July 2024; if they ever change
    # the name on the actual account, we can remove this fudge.
    if user_id == "62173425@N02" and realname == "tockholm Transport Museum":
        return "Stockholm Transport Museum"

    # This is a frequent commenter on Flickr Commons photos.  There's a
    # realname present if you visit their profile on Flickr.com,
    # but it isn't returned in the API.
    #
    # This name needed fixing on 7 August 2024 and I've reported it
    # as an API bug; if it ever gets fixed, we can remove this branch.
    if user_id == "32162360@N00" and username == "ɹǝqɯoɔɥɔɐǝq" and realname is None:
        return "beachcomber australia"

    return realname



"""
Methods for getting information about collections of photos in Flickr
(albums, galleries, groups, and so on).
"""

import typing
from xml.etree import ElementTree as ET

from nitrate.xml import find_optional_text, find_required_elem, find_required_text

from flickr_photos_api.date_parsers import parse_date_taken, parse_timestamp
from .license_methods import LicenseMethods
from .user_methods import UserMethods
from ..exceptions import ResourceNotFound
from ..types import (
    CollectionOfPhotos,
    GroupInfo,
    MediaType,
    PhotosInAlbum,
    PhotosInGallery,
    PhotosInGroup,
    SinglePhoto,
    User,
    create_user,
    get_machine_tags,
)
from ..utils import parse_location, parse_safety_level, parse_sizes


class CollectionMethods(LicenseMethods, UserMethods):
    """
    Methods for getting collections of photos from Flickr,
    e.g. albums or galleries.
    """

    def _from_collection_photo(
        self, photo_elem: ET.Element, owner: User | None
    ) -> SinglePhoto:
        """
        Given a <photo> element from a collection response, extract all the photo info.
        """
        photo_id = photo_elem.attrib["id"]

        secret = photo_elem.attrib["secret"]
        server = photo_elem.attrib["server"]
        farm = photo_elem.attrib["farm"]

        original_format = photo_elem.attrib.get("originalformat")

        if owner is None:
            owner = create_user(
                user_id=photo_elem.attrib["owner"],
                username=photo_elem.attrib["ownername"],
                realname=photo_elem.attrib.get("realname"),
                path_alias=photo_elem.attrib["pathalias"],
            )

        assert owner is not None

        safety_level = parse_safety_level(photo_elem.attrib["safety_level"])

        license = self.lookup_license_by_id(id=photo_elem.attrib["license"])

        title = photo_elem.attrib["title"] or None
        description = find_optional_text(photo_elem, path="description")
        tags = photo_elem.attrib["tags"].split()

        date_posted = parse_timestamp(photo_elem.attrib["dateupload"])
        date_taken = parse_date_taken(
            value=photo_elem.attrib["datetaken"],
            granularity=photo_elem.attrib["datetakengranularity"],
            unknown=photo_elem.attrib["datetakenunknown"] == "1",
        )

        # The lat/long/accuracy fields will always be populated, even
        # if there's no geo-information on this photo -- they're just
        # set to zeroes.
        #
        # We have to use the presence of geo permissions on the
        # <photo> element to determine if there's actually location
        # information here, or if we're getting the defaults.
        if photo_elem.attrib.get("geo_is_public") == "1":
            location = parse_location(photo_elem)
        else:
            location = None

        count_comments = int(photo_elem.attrib["count_comments"])
        count_views = int(photo_elem.attrib["count_views"])

        assert owner["photos_url"].endswith("/")
        url = owner["photos_url"] + photo_id + "/"

        sizes = parse_sizes(photo_elem)

        assert photo_elem.attrib["media"] in {"photo", "video"}
        media_type = typing.cast(MediaType, photo_elem.attrib["media"])

        return {
            "id": photo_id,
            "media": media_type,
            "secret": secret,
            "server": server,
            "farm": farm,
            "original_format": original_format,
            "owner": owner,
            "safety_level": safety_level,
            "license": license,
            "title": title,
            "description": description,
            "tags": tags,
            "machine_tags": get_machine_tags(tags),
            "date_posted": date_posted,
            "date_taken": date_taken,
            "location": location,
            "count_comments": count_comments,
            "count_views": count_views,
            "url": url,
            "sizes": sizes,
        }

    extras = [
        "license",
        "date_upload",
        "date_taken",
        "media",
        "original_format",
        "owner_name",
        "url_sq",
        "url_t",
        "url_s",
        "url_m",
        "url_o",
        "tags",
        "geo",
        # These parameters aren't documented, but they're quite
        # useful for our purposes!
        "url_q",  # Large Square
        "url_l",  # Large
        "description",
        "safety_level",
        "realname",
        "path_alias",
        "count_comments",
        "count_views",
    ]

    def _create_collection(
        self, collection_elem: ET.Element, owner: User | None = None
    ) -> CollectionOfPhotos:
        """
        This gets pagination information and extracts individual <photo>
        elements from a collection response.
        """
        photos = [
            self._from_collection_photo(photo_elem, owner=owner)
            for photo_elem in collection_elem.findall("photo")
        ]

        # The wrapper element includes a couple of attributes related
        # to pagination, e.g.
        #
        #     <photoset pages="1" total="2" …>
        #
        count_pages = int(collection_elem.attrib["pages"])
        count_photos = int(collection_elem.attrib["total"])

        return {
            "photos": photos,
            "count_pages": count_pages,
            "count_photos": count_photos,
        }

    def get_photos_in_album(
        self,
        album_id: str,
        user_id: str | None = None,
        user_url: str | None = None,
        page: int = 1,
        per_page: int = 10,
    ) -> PhotosInAlbum:
        """
        Get a page of photos from an album.

        You need to pass a numeric album ID and one of the ``user_id`` or ``user_url``.

        For example, if the album URL is

            https://www.flickr.com/photos/158685238@N03/albums/72177720313849533/

        then you need to pass one of:

            {"album_id": "72177720313849533", "user_id": "158685238@N03"}
            {"album_id": "72177720313849533", "user_url": "https://www.flickr.com/photos/158685238@N03/"}

        """
        user_id = self._ensure_user_id(user_id=user_id, user_url=user_url)

        return self._get_photos_in_album(
            user_id=user_id, album_id=album_id, page=page, per_page=per_page
        )

    def _get_photos_in_album(
        self, *, user_id: str, album_id: str, page: int, per_page: int
    ) -> PhotosInAlbum:
        """
        Get a page of photos from an album.
        """
        # https://www.flickr.com/services/api/flickr.photosets.getPhotos.html
        resp = self.call(
            method="flickr.photosets.getPhotos",
            params={
                "user_id": user_id,
                "photoset_id": album_id,
                "extras": ",".join(self.extras),
                "page": page,
                "per_page": per_page,
            },
            exceptions={
                "1": ResourceNotFound(f"Could not find album with ID: {album_id!r}"),
                "2": ResourceNotFound(f"Could not find user with ID: {user_id!r}"),
            },
        )

        # Albums are always non-empty, so we know we'll find something here
        photoset_elem = find_required_elem(resp, path="photoset")
        photo_elem = find_required_elem(photoset_elem, path="photo")

        owner = create_user(
            user_id=photoset_elem.attrib["owner"],
            username=photoset_elem.attrib["ownername"],
            realname=photo_elem.attrib.get("realname"),
            path_alias=photo_elem.attrib["pathalias"],
        )

        album_title = photoset_elem.attrib["title"]

        return {
            **self._create_collection(photoset_elem, owner=owner),
            "album": {
                "owner": owner,
                "title": album_title,
            },
        }

    def get_photos_in_gallery(
        self, *, gallery_id: str, page: int = 1, per_page: int = 10
    ) -> PhotosInGallery:
        """
        Get a page of photos in a gallery.
        """
        # https://www.flickr.com/services/api/flickr.galleries.getPhotos.html
        resp = self.call(
            method="flickr.galleries.getPhotos",
            params={
                "gallery_id": gallery_id,
                "get_gallery_info": "1",
                "extras": ",".join(self.extras),
                "page": page,
                "per_page": per_page,
            },
            exceptions={
                "1": ResourceNotFound(f"Could not find gallery with ID: {gallery_id!r}")
            },
        )

        gallery_elem = find_required_elem(resp, path="gallery")

        gallery_title = find_required_text(gallery_elem, path="title")
        gallery_owner_name = gallery_elem.attrib["username"]

        photos_elem = find_required_elem(resp, path="photos")

        return {
            **self._create_collection(photos_elem),
            "gallery": {"owner_name": gallery_owner_name, "title": gallery_title},
        }

    def get_photos_in_user_photostream(
        self,
        user_id: str | None = None,
        user_url: str | None = None,
        page: int = 1,
        per_page: int = 10,
    ) -> CollectionOfPhotos:
        """
        Get a page of photos from a user's photostream.

        You need to pass either the ``user_id`` or ``user_url``.

        For example, if the person's URL is

            https://www.flickr.com/photos/158685238@N03/

        then you need to pass one of:

            {"user_id": "158685238@N03"}
            {"user_url": "https://www.flickr.com/photos/158685238@N03/"}

        """
        user_id = self._ensure_user_id(user_id=user_id, user_url=user_url)

        # See https://www.flickr.com/services/api/flickr.people.getPublicPhotos.html
        resp = self.call(
            method="flickr.people.getPublicPhotos",
            params={
                "user_id": user_id,
                "extras": ",".join(self.extras),
                "page": page,
                "per_page": per_page,
            },
            exceptions={
                "1": ResourceNotFound(f"Could not find user with ID: {user_id!r}")
            },
        )

        first_photo = resp.find(".//photo")

        # The user hasn't uploaded any photos
        if first_photo is None:
            return {"count_pages": 1, "count_photos": 0, "photos": []}

        owner = create_user(
            user_id=first_photo.attrib["owner"],
            username=first_photo.attrib["ownername"],
            realname=first_photo.attrib.get("realname"),
            path_alias=first_photo.attrib["pathalias"],
        )

        photos_elem = find_required_elem(resp, path="photos")

        return self._create_collection(photos_elem, owner=owner)

    def _lookup_group_from_url(self, *, url: str) -> GroupInfo:
        """
        Given the link to a group's photos or profile, return some info.
        """
        # See https://www.flickr.com/services/api/flickr.urls.lookupGroup.html
        resp = self.call(
            method="flickr.urls.lookupGroup",
            params={"url": url},
            exceptions={
                "1": ResourceNotFound(f"Could not find group with URL: {url!r}")
            },
        )

        # The lookupUser response is of the form:
        #
        #       <group id="34427469792@N01">
        #         <groupname>FlickrCentral</groupname>
        #       </group>
        #
        group_elem = find_required_elem(resp, path=".//group")

        return {
            "id": group_elem.attrib["id"],
            "name": find_required_text(group_elem, path="groupname"),
        }

    def get_photos_in_group_pool(
        self, *, group_url: str, page: int = 1, per_page: int = 10
    ) -> PhotosInGroup:
        """
        Get a page of photos in a group pool.
        """
        group_info = self._lookup_group_from_url(url=group_url)

        # See https://www.flickr.com/services/api/flickr.groups.pools.getPhotos.html
        resp = self.call(
            method="flickr.groups.pools.getPhotos",
            params={
                "group_id": group_info["id"],
                "extras": ",".join(self.extras),
                "page": page,
                "per_page": per_page,
            },
        )

        photos_elem = find_required_elem(resp, path="photos")

        return {
            **self._create_collection(photos_elem),
            "group": group_info,
        }

    def get_photos_with_tag(
        self, *, tag: str, page: int = 1, per_page: int = 10
    ) -> CollectionOfPhotos:
        """
        Get a page of photos in a tag.

        Note that tag pagination and ordering results can be inconsistent,
        especially for large tags -- it's tricky to do an "exhaustive" search
        of a Flickr tag.
        """
        resp = self.call(
            method="flickr.photos.search",
            params={
                "tags": tag,
                "page": page,
                "per_page": per_page,
                # This is so we get the same photos as you see on the "tag" page
                # under "All Photos Tagged XYZ" -- if you click the URL to the
                # full search results, you end up on a page like:
                #
                #     https://flickr.com/search/?sort=interestingness-desc&…
                #
                "sort": "interestingness-desc",
                "extras": ",".join(self.extras),
            },
        )

        photos_elem = find_required_elem(resp, path="photos")

        return self._create_collection(photos_elem)



"""
On Flickr, machine tag are used to store structured data in types,
e.g. ``geo:lat=12.34`` is a tag with location information.

This file has a few functions for dealing with machine tags.
"""

import collections
import re
import typing


# The best documentation I can find for Flickr's implementation of
# machine tags is a group post from a Flickr staff member in 2007:
# https://www.flickr.com/groups/51035612836@N01/discuss/72157594497877875/
#
# A machine tag is made of three parts:
#
#     {namespace}:{predicate}={value}
#
# The ``namespace`` and ``predicate`` can only contain ASCII characters.
MACHINE_TAG_RE = re.compile(
    r"""
    ^
    (?P<namespace>[a-zA-Z0-9_]+)
    :
    (?P<predicate>[a-zA-Z0-9_]+)
    =
    (?P<value>.+)
    $
    """,
    re.VERBOSE,
)


MachineTags: typing.TypeAlias = dict[str, list[str]]


def get_machine_tags(tags: list[str]) -> MachineTags:
    """
    Given a list of raw tags on Flickr, parse the machine tags
    as key-value pairs.

    The keys are namespace/predicate; the values are values.

    Example:

        >>> tags = [
        ...     "square",
        ...     "shape:sides=4",
        ...     "shape:color=red",
        ...     "shape:color=blue"
        ... ]
        ...
        >>> get_machine_tags(tags)
        {"shape:sides": ["4"], "shape:color": ["red", "blue"]}

    This function is a "best effort" parsing of machine tags -- it may
    not match Flickr perfectly, but is meant to make it easier for
    callers to work with machine tags.
    """
    result: MachineTags = collections.defaultdict(list)

    for t in tags:
        if m := MACHINE_TAG_RE.match(t):
            namespace = m.group("namespace")
            predicate = m.group("predicate")
            value = m.group("value")

            result[f"{namespace}:{predicate}"].append(value)

    return dict(result)
